import os
import sys
import re
import math

folder_name = sys.argv[1]
num_results = sys.argv[2]
query = sys.argv[3]

filenames = os.listdir(folder_name)
filenames.sort(key=lambda f: int(re.sub(r'\D', '', f)))

inv_ind = {}

def create_ii():
    for fname in filenames:
        doc_num = int(re.sub(r'\D', '', fname))
        with open(f'{folder_name}/{fname}') as f:
            read_data = f.read()
            read_data = read_data.lower()
            read_data = re.sub(r'\W', ' ', read_data)
            read_data = read_data.split()
            for word_index in range(len(read_data)):
                word = read_data[word_index]
                if word in inv_ind:
                    inv_ind[word].append((doc_num, word_index))
                else:
                    inv_ind[word] = [(doc_num, word_index)]


def print_ii():
    for key in sorted(inv_ind):
        print(key)
        s = set()
        for (a,b) in inv_ind[key]:
            s.add(a)
        print(len(s), end=",")
        print(len(inv_ind[key]), end=",")
        print(','.join(str(tup) for tup in inv_ind[key]))


def first(t):
    return inv_ind[t][0]


def last(t):
    return inv_ind[t][-1]


def prev(t, docId, current):
    jump = 1
    low = 0
    high = low + jump
    if t not in inv_ind.keys():
        return (-math.inf, -math.inf)

    if (compare(inv_ind[t][-1], docId, current) == -1):
       return inv_ind[t][-1]

    if (compare(inv_ind[t][0], docId, current) == 0 or
        compare(inv_ind[t][0], docId, current) == 1):
        return (-math.inf, -math.inf)


    while (high < len(inv_ind[t]) - 1 and 
           compare(inv_ind[t][high], docId, current) == -1):
        low = high
        jump *= 2 
        high = low + jump

    if high > len(inv_ind[t]):
        high = len(inv_ind[t]) - 1
    pos = binary_search_prev(t, docId, low, high, current)
    return inv_ind[t][pos]


def binary_search_prev(t, docId, low, high, current):
    while high-low > 1:
        mid = (low+high)//2
        if compare(inv_ind[t][mid], docId, current) == -1:
            low = mid
        else:
            high = mid
    return low


def next(t, docId, current):
    jump = 1
    low = 0
    high = low + jump
    if t not in inv_ind.keys():
        return (math.inf, math.inf)

    if (compare(inv_ind[t][-1], docId, current) == -1 or
       compare(inv_ind[t][-1], docId, current) == 0):
       return (math.inf, math.inf)

    if compare(inv_ind[t][0], docId, current) == 1:
        return inv_ind[t][0]


    while (high < len(inv_ind[t]) - 1 and 
           (compare(inv_ind[t][high], docId, current) == -1 or
            compare(inv_ind[t][high], docId, current) == 0)):
        low = high
        jump *= 2 
        high = low + jump

    if high > len(inv_ind[t]):
        high = len(inv_ind[t]) - 1
    pos = binary_search_next(t, docId, low, high, current)
    return inv_ind[t][pos]


def binary_search_next(t, docId, low, high, current):
    while high-low > 1:
        mid = (low+high)//2
        if(compare(inv_ind[t][mid], docId, current) == -1 or
           compare(inv_ind[t][mid], docId, current) == 0):
            low = mid
        else:
            high = mid
    return high


def nextPhrase(t_list, docId, position):
    v = (docId, position)
    n = len(t_list)
    for i in range(n):
        v = next(t_list[i], v[0], v[1])
    if (v[0] == math.inf):
        return [(math.inf, math.inf), (math.inf, math.inf)]
    u = v
    for i in range(n-2, -1, -1):
        u = prev(t_list[i], u[0], u[1])
    if v[0] == u[0] and v[1]-u[1] == n-1:
        return [u,v]
    else:
        return nextPhrase(t_list, u[0], u[1])


def prevPhrase(t_list, docId, position):
    v = (docId, position)
    n = len(t_list)
    for i in range(n-1, -1, -1):
        v = prev(t_list[i], v[0], v[1])
    if (v[0] == -math.inf):
        return [(-math.inf, -math.inf), (-math.inf, -math.inf)]
    u = v
    for i in range(1, n):
        u = next(t_list[i], u[0], u[1])
    if v[0] == u[0] and u[1]-v[1] == n-1:
        return [v,u]
    else:
        return prevPhrase(t_list, u[0], u[1])


def docRight(q, docId, pos):
    stack = []

    q = q.split(' ')
    q.reverse()

    for val in q:
        if val in ['_AND', '_OR']:
            op1 = stack.pop()
            op2 = stack.pop()
            if val=='_AND':
                result = max_t(op1[0], op1[1], op2[0], op2[1])
            if val=='_OR':
                result = min_t(op1[0], op1[1], op2[0], op2[1])
            stack.append(result)
        else:
            val = val.lower()
            t_list = val.split('_')                
            result = nextDoc(t_list, docId, pos)
            stack.append(result)

    return stack.pop()


def docLeft(q, docId, pos):
    stack = []

    q = q.split(' ')
    q.reverse()

    for val in q:
        if val in ['_AND', '_OR']:
            op1 = stack.pop()
            op2 = stack.pop()
            if val=='_OR':
                result = max_t(op1[0], op1[1], op2[0], op2[1])
            if val=='_AND':
                result = min_t(op1[0], op1[1], op2[0], op2[1])
            stack.append(result)
        else:
            val = val.lower()
            t_list = val.split('_')                
            result = prevDoc(t_list, docId, pos)
            stack.append(result)

    return stack.pop()


def max_t(docId1, pos1, docId2, pos2):
    if (compare((docId1, pos1), docId2, pos2) == -1):
        return (docId2, pos2)
    else:
        return (docId1, pos1)


def min_t(docId1, pos1, docId2, pos2):
    if (compare((docId1, pos1), docId2, pos2) == -1):
        return (docId1, pos1)
    else:
    
        return (docId2, pos2)

def nextDoc(t, docId, pos):
    result = nextPhrase(t, docId, pos)
    return result[1]


def prevDoc(t, docId, pos):
    result = prevPhrase(t, docId, pos)
    return result[0]


def nextSolution(q, docId, pos):
    v = docRight(q, docId, pos)
    if v == (math.inf, math.inf):
        return [(math.inf, math.inf), (math.inf, math.inf)]
    u = docLeft(q, v[0], v[1] + 1)
    if (u[0] == v[0]):
        return [u,v]
    else:
        return nextSolution(q, v[0], v[1]-1)


def rankProximity(q, k):
    result = []
    docIdStart = -math.inf
    posStart = -math.inf

    cover = nextSolution(q, docIdStart, posStart)
    prevDoc = cover[0][0]
    docIdStart = cover[0][0]
    posStart = cover[0][1]
    docIdEnd = cover[1][0]
    posEnd = cover[1][1]
    score = 0

    while (posStart < math.inf and docIdStart < math.inf):
        if (prevDoc < docIdStart):
            result.append((prevDoc, score))
            prevDoc = docIdStart
            score = 0
        score += 1/(posEnd - posStart + 1)
        cover = nextSolution(q, docIdStart, posStart)
        docIdStart = cover[0][0]
        posStart = cover[0][1]
        docIdEnd = cover[1][0]
        posEnd = cover[1][1]
    
    if (prevDoc < math.inf):
        result.append((prevDoc, score))
    
    result.sort(reverse=True, key=lambda tup: tup[1])
    return result[0:k]


def compare(tup, docId2, current2):
    docId1, current1 = tup
    if docId1 < docId2:
        return -1
    elif docId1 > docId2:
        return 1
    elif current1 < current2:
        return -1
    elif current1 > current2:
        return 1
    else:
        return 0


def print_next(term, doc_num, pos):
    print(f'next("{term}", {doc_num}, {pos}) = {next(term, doc_num, pos)}')


def print_prev(term, doc_num, pos):
    print(f'prev("{term}", {doc_num}, {pos}) = {prev(term, doc_num, pos)}')

def print_result(result):
    print(f'DocId Score')
    for i in result:
        print(f'{i[0]} {i[1]}')

def main():
    create_ii()
    # print_ii()

    output = rankProximity(query, int(num_results))
    print_result(output)

    # print("Position of 'dog': " + str(first("dog")))
    # print("Position of 'dog': " + str(last("dog")))
    # print_next('dog', 1, -math.inf)
    # print_next('dog', 1, 0)
    # print_next('dog', 1, 1)
    # print_next('dog', 1, 5)
    # print_next('dog', 1, 6)
    # print_next('dog', 1, 7)
    # print_next('dog', 1, 8)
    # print_next('dog', 2, 1)
    # print_next('dog', 2, 9)
    # print_next('dog', 2, 14)
    # print_next('dog', 3, 0)
    # print_next('dog', 3, 2)
    # print_next('dog', 4, 0)
    # print_next('dog', 4, 3)
    # print_next('dog', 4, 7)
    # print_next('dog', 4, 8)


    # print('='*20)
    # print_prev('dog', 1, 0)
    # print_prev('dog', 1, 1)
    # print_prev('dog', 1, 5)
    # print_prev('dog', 1, 6)
    # print_prev('dog', 1, 7)
    # print_prev('dog', 1, 8)
    # print_prev('dog', 2, 1)
    # print_prev('dog', 2, 9)
    # print_prev('dog', 2, 14)
    # print_prev('dog', 3, 0)
    # print_prev('dog', 3, 2)
    # print_prev('dog', 4, 0)
    # print_prev('dog', 4, 3)
    # print_prev('dog', 4, 7)
    # print_prev('dog', 4, 8)

    # print(nextPhrase(['lazy', 'dog'], 1, 23))
    # print(nextPhrase(['brown', 'dog'], 1, 23))
    # print(nextPhrase(['brown', 'dog'], 3, 23))
    # print(nextPhrase(['dog', 'hawaiian'], 1, 3))
    # print(nextPhrase(['the3', 'dog'], -math.inf, -math.inf))
    # print(nextPhrase(['dog', 'dog'], 4, 0))
    # print(nextPhrase(['dog'], -1, -1))
    
    # print('='*20)

    # print(prevPhrase(['lazy', 'dog'], 1, 23))
    # print(prevPhrase(['brown', 'dog'], 2, 23))
    # print(prevPhrase(['brown', 'dog'], 3, 23))
    # print(prevPhrase(['dog', 'hawaiian'], 3, 3))
    # print(prevPhrase(['the3', 'dog'], 3, 3))
    # print(prevPhrase(['dog', 'dog'], 6, 3))

    # print(nextDoc(['dog', 'dog'], 4, 0))
    # print(nextDoc(['dog'], -1, -1))
    # print(nextDoc(['lazy', 'dog'], 1, 2))

    # print('='*20)

    # print(prevDoc(['the3', 'dog'], 3, 3))
    # print(prevDoc(['dog'], 6, 3))
    # print(prevDoc(['only', 'dogs'], 3, 4))

    # print('='*20)

    # print(docRight('_OR big_lazy fence_dog', 0, 0))
    # print(docRight('_AND big_lazy fence_dog', 0, 0))

    # print(docLeft('_OR big_lazy fence_dog', 6, 0))
    # print(docLeft('_AND big_lazy fence_dog', 6, 0))

    # print(docRight('_AND _OR idk fence_dog new_line', 0, 0))

    # print(docRight('_OR good_dog _AND bad cat', 5, -1))
    # print(docLeft('_OR good_dog _AND bad cat', 5, 3))
    # print(docRight('_OR good_dog _AND bad cat', 5, 3))
    # print(docLeft('_OR good_dog _AND bad cat', 5, 100))
    # print(nextSolution('_OR good_dog _AND bad cat', 5, -math.inf))
    # print('='*20)
    # print(rankProximity('_OR good_dog _AND bad cat', 5))
    # print('='*20)
    # print(rankProximity('_OR lazy_dog _AND cat dog', 5))
    


    
    




if __name__ == "__main__":
    main()


# def docRight(Q,u):
#     nextDoc(Q,u)

# def nextDoc(t,u):
#         read_data = f.read()
#         read_data = read_data.lower()

# folder_name = sys.argv[1]
# query = sys.argv[2]